/******************************************************************************
 *                                                                            *
 * /sbin/hotplug kernel helper application for hotplugabled devices           *
 * (C) 2008 Coolstream International Limited                                  *
 *                                                                            *
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/stat.h>
#include <blkid/blkid.h>
#include <mntent.h>

#define DEBUG

int FIFO_usable = 0;

#define FIRMWARE_PATH   "/lib/firmware"

void log_err(char *firmware, int err)
{
	char *driver;

	driver = getenv("PHYSDEVDRIVER");
	if (!driver)
		driver = "(unknown)";
	openlog("firmware_helper", LOG_PID, LOG_USER);
	syslog(LOG_ERR, "Loading of %s for %s driver failed: %s", firmware, driver, strerror(err));
	closelog();
}

/* Set the 'loading' attribute for a firmware device.
 * 1 == currently loading
 * 0 == done loading
 * -1 == error
 */
int set_loading(char *device, int value)
{
	char *path;
	int rc;
	FILE *f;

	size_t path_size = strlen(device) + 15;
	path = alloca(path_size);
	snprintf(path,path_size, "/sys/%s/loading", device);
	f = fopen(path, "w");
	if (!f)
		return 1;
	rc = fprintf(f, "%d", value);
	fclose(f);
	if (rc < 0)
		return rc;
	return 0;
}

int load_firmware(void)
{
	char *devpath, *firmware, *action;
	struct stat sbuf;
	int fw_fd, count;
	int rc;
	char *fw_path, *data_path;
	char *fw_buffer;

	devpath = getenv("DEVPATH");
	if (!devpath)
		return 0;

	firmware = getenv("FIRMWARE");
	if (!firmware)
		return 0;

	action = getenv("ACTION");
	if (strcmp(action, "add"))
		return 0;

	set_loading(devpath, 1);

	size_t fw_path_size = strlen(FIRMWARE_PATH) + strlen(firmware) + 2;
	fw_path = malloc(fw_path_size);
	if (!fw_path) {
		rc = errno;
		goto out_err;
	}

	snprintf(fw_path,fw_path_size, "%s/%s", FIRMWARE_PATH, firmware);
	fw_fd = open(fw_path, O_RDONLY);
	if (fw_fd == -1) {
		rc = errno;
		goto out_err;
	}

	fstat(fw_fd, &sbuf);
	fw_buffer = malloc(sbuf.st_size);
	if (!fw_buffer) {
		rc = errno;
		goto out_err;
	}
	if (read(fw_fd, fw_buffer, sbuf.st_size) != sbuf.st_size) {
		rc = errno;
		free(fw_buffer);
		goto out_err;
	}
	close(fw_fd);

	size_t data_path_size = strlen(devpath) + 12;
	data_path = malloc(data_path_size);
	if (!data_path) {
		rc = errno;
		free(fw_buffer);
		goto out_err;
	}

	snprintf(data_path,data_path_size, "/sys/%s/data", devpath);

	fw_fd = open(data_path, O_RDWR);
	if (fw_fd == -1) {
		rc = errno;
		free(data_path);
		free(fw_buffer);
		goto out_err;
	}

	count = 0;
	while (count < sbuf.st_size) {
		int c;

		c = write(fw_fd, fw_buffer + count, sbuf.st_size - count);
		if (c <= 0) {
			rc = errno;
			free(data_path);
			free(fw_buffer);
			goto out_err;
		}
		count += c;
	}
	close(fw_fd);
	free(fw_buffer);
	free(data_path);
	set_loading(devpath, 0);
	return 0;
out_err:
	if (fw_fd != -1)
		close(fw_fd);
	set_loading(devpath, -1);
	log_err(fw_path, rc);
	return rc;
}

/******************************************************************************/

static void _get_volume_label(char *dev, char *label, long unsigned int label_size)
{
	static blkid_probe blprobe;
	char devpath[11];
	const char *lbl = NULL;
	int fd, prv = 1;

	snprintf(devpath,sizeof(devpath), "/dev/%s%c", dev, '\0');

	fd = open(devpath, O_RDONLY);
	if (fd >= 0)
	{
		blprobe = blkid_new_probe();
		if (blprobe)
		{
			if (!blkid_probe_set_device(blprobe, fd, 0, 0))
			{
				blkid_probe_enable_superblocks(blprobe, 1);
				blkid_probe_set_superblocks_flags(blprobe, BLKID_SUBLKS_LABEL);

				if (!blkid_do_safeprobe(blprobe))
				{
					prv = blkid_probe_lookup_value(blprobe, "LABEL", &lbl, NULL);
				}
			}
			blkid_free_probe(blprobe);
		}
		close(fd);
	}


	if (prv == 0)
	{
		snprintf(label,label_size, "/media/%s%c", lbl, '\0');
	}
	else
	{
		snprintf(label,label_size, "/media/%s%c", dev, '\0');
	}
}

/******************************************************************************/

static void _get_mount_path(char *dev, char *mntpth, long unsigned int mntpth_size)
{
	struct mntent *mtabent;
	char devpath[11];
	FILE *fp = NULL;
	int crv = 1;

	snprintf(devpath,sizeof(devpath), "/dev/%s%c", dev, '\0');

	fp = setmntent("/proc/mounts", "r");
	if (fp != NULL)
	{
		while ((mtabent = getmntent(fp)))
		{
			if ((crv = strncmp(mtabent->mnt_fsname, devpath, strlen(devpath))) == 0)
			{
				strcpy(mntpth, (char *) mtabent->mnt_dir);
				break;
			}
		}
		endmntent(fp);
	}

	if (crv != 0)
	{
		snprintf(mntpth,mntpth_size, "/media/%s%c", dev, '\0');
	}
}

/******************************************************************************/

// static void _make_vendor_model_mountpath(unsigned char *prefix, unsigned char *str)
// {
// 	unsigned char infbuf[1024];
// 	unsigned char vendor[1024];
// 	unsigned char model[1024];
// 	int x;
// 
// 	memset(vendor, 0, 1024);
// 	memset(model, 0, 1024);
// 
// 	sprintf(infbuf, "/sys%s/vendor%c", getenv("PHYSDEVPATH"), '\0');
// 
// 	FILE *inf = fopen(infbuf, "r");
// 	if (inf)
// 	{
// 		fgets(infbuf, 1024, inf);
// 		fclose (inf);
// 		inf = NULL;
// 		if ((infbuf[0] < 0x21) || (infbuf[0] > 0x7F))
// 			sprintf(infbuf, "unknown vendor%c", '\0');
// 
// 		/* remove spaces at the end */
// 		int z = strlen(infbuf) - 1;
// 		if (z > 0)
// 		{
// 			while ((z > 0) && (infbuf[z] < 0x21))
// 				z--;
// 		}
// 		z++;
// 
// 		/* replace "/" and character smaller 0x20 by spaces, busybox and samba can not handle this */
// 		for (x = 0; x < z; x++)
// 			if ((infbuf[x] == '/') || (infbuf[x] < 0x20) || (infbuf[x] > 0x7F))
// 				infbuf[x] = ' ';
// 
// 		/* remove that stupid CR */
// 		strncpy(vendor, infbuf, z);
// 	}
// 
// 	sprintf(infbuf, "/sys%s/model%c", getenv("PHYSDEVPATH"), '\0');
// 
// 	inf = fopen(infbuf, "r");
// 	if (inf)
// 	{
// 		fgets(infbuf, 1024, inf);
// 		fclose (inf);
// 		inf = NULL;
// 		if ((infbuf[0] < 0x21) || (infbuf[0] > 0x7F))
// 			sprintf(infbuf, "unknown model%c", '\0');
// 
// 		/* remove spaces */
// 		int z = strlen(infbuf) - 1;
// 		if (z > 0)
// 		{
// 			while ((z > 0) && (infbuf[z] < 0x21))
// 				z--;
// 		}
// 		z++;
// 
// 		/* replace "/" and character smaller 0x20 by spaces, busybox and samba can not handle this */
// 		for (x = 0; x < z; x++)
// 			if ((infbuf[x] == '/') || (infbuf[x] < 0x20) || (infbuf[x] > 0x7F))
// 				infbuf[x] = ' ';
// 
// 		/* remove that stupid CR */
// 		strncpy(model, infbuf, z);
// 	}
// 	//snprintf(str,sizeof(str), "/dev/shm/usb/%s - %s - %s%c", prefix, vendor, model, '\0');
// 	snprintf(str,sizeof(str), "/media/%s", prefix, '\0');
// }

/******************************************************************************/

static void _write_fifo(unsigned char action, unsigned char source, char *data)
{
	if (!FIFO_usable)
		return;

	unsigned char str[1024];
	FILE *fifo = fopen("/tmp/hotplug", "w");
	if (fifo)
	{
		memset(str, 0, 1024);
		fprintf(fifo, "%c%c%c%s%c%c", action, source, '"', data, '"', '\0');
		fclose(fifo);
	}
}

/******************************************************************************/

static void _write_fifo2(unsigned char action, unsigned char source, char *data)
{
	if (!FIFO_usable)
		return;

	unsigned char str[1024];
	FILE *fifo = fopen("/tmp/hotplug", "w");
	if (fifo)
	{
		memset(str, 0, 1024);
		fprintf(fifo, "%c%c%s%c", action, source, data, '\0');
		fclose(fifo);
	}
}

/******************************************************************************/

int main(int argc, char *argv[]/*, char *envp[]*/)
{
	if ((!argv[1]) || (!getenv("ACTION")))
	{
		printf("you can not run this program without knowledge about it.\nIt's executed directly by the Linux-Kernel.%i\n",argc);
		return 0;
	}

	int fd = open("/tmp/hotplug", O_WRONLY | O_NONBLOCK);
	if (fd != -1)
	{
		FIFO_usable = 1;
		close(fd);
	}
	else
		FIFO_usable = 0;

	unsigned char addc;
	if (strncmp(getenv("ACTION"), "add", 3) == 0)
		addc = 'A';
	else
		addc = 'R';

#ifdef DEBUG
	FILE *dbg = fopen("/dev/console", "w");
	fprintf(dbg, "HOTPLUG: source: %s, action: %s, product: %s, interface: %s\n", argv[1], getenv("ACTION"), getenv("PRODUCT"), getenv("INTERFACE"));
#endif

	if (strncmp(argv[1], "usb", 3) == 0)
	{
		if (strncmp(getenv("ACTION"), "add", 3) == 0)
		{
			int ifclass, ifsubclass, ifproto;
			unsigned int idvendor, idproduct, bcddevice;
//			unsigned char buf[1024];
			if (sscanf(getenv("INTERFACE"), "%d/%d/%d", &ifclass, &ifsubclass, &ifproto) == 3)
			{

#ifdef DEBUG
				fprintf(dbg, "HOTPLUG: ifclass/ifsubclass/ifproto: %d/%d/%d\n", ifclass, ifsubclass, ifproto);
#endif
				int load_rtdriver = 0;
				switch (ifclass)
				{
				case 3:	/* input */
					/* do nothing. usb-hid is inside the kernel, but inform the app to show some messages */
					switch (ifsubclass)
					{
					case 0:	/* Keyboard */
						_write_fifo2(addc, 'I', "Keyboard");
						break;
					case 1:	/* Mouse */
						_write_fifo2(addc, 'I', "Mouse");
						break;
					default:	/* unknown, so use HID */
						_write_fifo2(addc, 'I', "HID");
						break;
					}
					break;
				case 8:	/* storage */
					_write_fifo2(addc, 'I', "Storage");
					/* load usb-storage module */
#ifdef DEBUG
					fprintf(dbg, "HOTPLUG: STORAGE...\n");
#endif
					//system("modprobe usb-storage 2>/dev/null");
					break;
				case 255:	/* vendor specific devices like the ASUS WLAN-adapter */
					/* Devices supported by our driver
					    Manufactor	Device	Name
					    0x0707	0xee13	SMC EZ-Connect
					    0x0b05	0x1706	ASUS WL-167G
					    0x0b05	0x1707	ASUS ???
					    0x13b1	0x000d	Cisco-Linksys ???
					    0x13b1	0x0011	Cisco-Linksys ???
					    0x148f	0x2570	Ralink RT2570 (base identification)
					    0x2001	0x3c00	D-Link DWL-G122
					*/
					if (sscanf(getenv("PRODUCT"), "%x/%x/%x", &idvendor, &idproduct, &bcddevice) == 3)
					{
						if ((idvendor == 0x0707) && (idproduct == 0xEE13))
						{
							_write_fifo2(addc, 'I', "SMC EZ-Connect WLAN-Adapter");
							load_rtdriver = 1;
						}
						if ((idvendor == 0x0B05) && (idproduct == 0x1706))
						{
							_write_fifo2(addc, 'I', "ASUS WL-167G WLAN-Adapter");
							load_rtdriver = 1;
						}
						if ((idvendor == 0x0B05) && (idproduct == 0x1707))
						{
							_write_fifo2(addc, 'I', "ASUS WLAN-Adapter");
							load_rtdriver = 1;
						}
						if ((idvendor == 0x13B1) && (idproduct == 0x000D))
						{
							_write_fifo2(addc, 'I', "Linksys WUSB54Gv4 WLAN-Adapter");
							load_rtdriver = 1;
						}
						if ((idvendor == 0x13B1) && (idproduct == 0x0011))
						{
							_write_fifo2(addc, 'I', "Linksys WUSB54GPv4 WLAN-Adapter");
							load_rtdriver = 1;
						}
						if ((idvendor == 0x148F) && (idproduct == 0x2570))
						{
							_write_fifo2(addc, 'I', "Ralink RT2570 WLAN-Adapter");
							load_rtdriver = 1;
						}
						if ((idvendor == 0x2001) && (idproduct == 0x3C00))
						{
							_write_fifo2(addc, 'I', "D-Link DWL-G122 Rev.B WLAN-Adapter");
							load_rtdriver = 1;
						}
						if (load_rtdriver)
							system("modprobe rt2570 ");
					}
					break;
				default:
					break;
				}
			}
		}
		else
		{
			int ifclass, ifsubclass, ifproto;
//			int idvendor, idproduct, bcddevice;
//			unsigned char buf[1024];
			if (sscanf(getenv("INTERFACE"), "%d/%d/%d", &ifclass, &ifsubclass, &ifproto) == 3)
			{
				switch (ifclass)
				{
				case 3:	/* input */
					/* do nothing. usb-hid is inside the kernel, but inform the app to show some messages */
					switch (ifsubclass)
					{
					case 0:	/* Keyboard */
						_write_fifo2(addc, 'I', "Keyboard");
						break;
					case 1:	/* Mouse */
						_write_fifo2(addc, 'I', "Mouse");
						break;
					default:	/* unknown, so use HID */
						_write_fifo2(addc, 'I', "HID");
						break;
					}
					break;
				case 8:	/* storage */
					_write_fifo2(addc, 'I', "Storage");
					break;
				default:
					break;
				}
			}
		}
	}
	else if (strncmp(argv[1], "block", 5) == 0)
	{
		char *pointer = getenv("DEVPATH");
		char *buf = NULL;
		if(pointer != NULL)
			buf = strdup(pointer);
		if (buf)
		{
			int len = strlen(buf);
			int pos = len - 1;
			while ((pos > 0) && (buf[pos] != '/'))
				pos--;
			pos++;
			char str[1024];
			char infbuf[1024];

			if (strncmp(getenv("ACTION"), "add", 3) == 0)
			{
				if ((buf[pos] == 's') && (buf[pos + 1] == 'd'))
				{
					/* create the device */
					snprintf(str,sizeof(str), "mknod /dev/%s block %s %s%c", &buf[pos], getenv("MAJOR"), getenv("MINOR"), '\0');
#ifdef DEBUG
					fprintf(dbg, "HOTPLUG: %s ...\n", str);
#endif
					system(str);
					/* sleep a little bit. The kernel is not so fast in creating the files we wanna read */
					usleep(5000000);
					/* make a mount-point */
					//_make_vendor_model_mountpath(&buf[pos], infbuf);
					_get_volume_label(&buf[pos], infbuf, sizeof(infbuf));
					snprintf(str,sizeof(str), "mkdir -p %c%s%c%c", '"', infbuf, '"', '\0');
#ifdef DEBUG
					fprintf(dbg, "HOTPLUG: %s ...\n", str);
#endif
					// usleep(600000);
					system(str);
					/* try to mount */
					snprintf(str,sizeof(str), "mount -t auto /dev/%s %c%s%c%c", &buf[pos], '"', infbuf, '"', '\0');
#ifdef DEBUG
					fprintf(dbg, "HOTPLUG:test auto mounting %s ...\n", str);
#endif
					int ret = system(str);
					if (ret)
					{
						//ntfs mount
						snprintf(str,sizeof(str), " ntfs-3g /dev/%s %c%s%c%c ", &buf[pos], '"', infbuf, '"', '\0');
						ret = system(str);
						snprintf(str,sizeof(str), "mount -t auto /dev/%s %c%s%c%c", &buf[pos], '"', infbuf, '"', '\0');
						usleep(300000);
					}
					if (ret)
					{
#ifdef DEBUG
						fprintf(dbg, "HOTPLUG: %s failed: %d\n", str, ret);
#endif
						/* mount failed due to missing media */
						snprintf(str,sizeof(str), "rmdir %c%s%c%c", '"', infbuf, '"', '\0');
#ifdef DEBUG
						fprintf(dbg, "HOTPLUG: %s ...\n", str);
#endif
						system(str);
					}
					else
					{
						/* put notification data into the FIFO /dev/shm/hotplug */
						_write_fifo2(addc, 'I', "Flash Media");
						_write_fifo('A', 'S', infbuf);
					}
				}
			}
			else
			{
				snprintf(str,sizeof(str), "%s", &buf[pos]);
				if ((strlen(str) > 3) && (buf[pos] == 's') && (buf[pos + 1] == 'd'))
				{
					usleep(400000);
					//_make_vendor_model_mountpath(&buf[pos], infbuf);
					_get_mount_path(&buf[pos], infbuf,sizeof(infbuf));
					_write_fifo2(addc, 'I', "Flash Media");
					snprintf(str,sizeof(str), "umount -lf %s%c", infbuf, '\0');
					if (!system(str))
					{
						snprintf(str,sizeof(str), "rm -rf %c%s%c%c", '"', infbuf, '"', '\0');
						system(str);
						snprintf(str,sizeof(str), "rm /dev/%s%c", &buf[pos], '\0');
						system(str);
					}
#ifdef DEBUG
					else
						fprintf(dbg, "HOTPLUG: unmounting /dev/%s failed\n", &buf[pos]);
#endif
					/* put notification data into the FIFO /dev/shm/hotplug */
#ifdef DEBUG
					fprintf(dbg, "HOTPLUG: emitting event in FIFO\n");
#endif
					_write_fifo('R', 'S', infbuf);
				}
			}
			free(buf);
		}
	}
	else if (strncmp(argv[1], "net", 3) == 0)
	{
		char *pointer = getenv("DEVPATH");
		char *buf = NULL;
		if(pointer != NULL)
			buf = strdup(pointer);

		if (buf)
		{
			int len = strlen(buf);
			int pos = len - 1;
			while ((pos > 0) && (buf[pos] != '/'))
				pos--;
			pos++;
//			unsigned char str[1024];
//			unsigned char infbuf[1024];

			/* put notification data into the FIFO /dev/shm/hotplug */
			_write_fifo2(addc, 'I', "Network");
			_write_fifo(addc, 'N', &buf[pos]);
		}
	}
	else if (strncmp(argv[1], "firmware", 8) == 0)
	{
		load_firmware();
	}

#ifdef DEBUG
	fclose(dbg);
#endif
	return 0;
}
